@@ -2,7 +2,7 @@
  * ProFTPD - FTP server daemon
  * Copyright (c) 1997, 1998 Public Flood Software
  * Copyright (c) 1999, 2000 MacGyver aka Habeeb J. Dihu <macgyver@tos.net>
- * Copyright (c) 2001-2009 The ProFTPD Project team
+ * Copyright (c) 2001-2016 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -16,7 +16,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
  *
  * As a special exemption, Public Flood Software/MacGyver aka Habeeb J. Dihu
  * and other respective copyright holders give permission to link this program
@@ -24,9 +24,7 @@
  * the source code for OpenSSL in the source distribution.
  */
 
-/* Authentication front-end for ProFTPD
- * $Id: auth.c,v 1.80.2.1 2010/08/30 17:37:39 castaglia Exp $
- */
+/* Authentication front-end for ProFTPD */
 
 #include "conf.h"
 #include "privs.h"
@@ -91,14 +89,16 @@ static void uidcache_create(void) {
 
     uid_tab = pr_table_alloc(auth_pool, 0);
 
-    if (pr_table_ctl(uid_tab, PR_TABLE_CTL_SET_KEY_CMP, uid_keycmp_cb) < 0) {
+    if (pr_table_ctl(uid_tab, PR_TABLE_CTL_SET_KEY_CMP,
+        (void *) uid_keycmp_cb) < 0) {
       pr_trace_msg(trace_channel, 2,
         "error setting key comparison callback for uidcache: %s",
         strerror(errno));
       ok = FALSE;
     }
 
-    if (pr_table_ctl(uid_tab, PR_TABLE_CTL_SET_KEY_HASH, uid_hash_cb) < 0) {
+    if (pr_table_ctl(uid_tab, PR_TABLE_CTL_SET_KEY_HASH,
+        (void *) uid_hash_cb) < 0) {
       pr_trace_msg(trace_channel, 2,
         "error setting key hash callback for uidcache: %s",
         strerror(errno));
@@ -115,6 +115,10 @@ static void uidcache_create(void) {
 }
 
 static void uidcache_add(uid_t uid, const char *name) {
+  if (!(auth_caching & PR_AUTH_CACHE_FL_UID2NAME)) {
+    return;
+  }
+
   uidcache_create();
 
   if (uid_tab) {
@@ -155,14 +159,16 @@ static void gidcache_create(void) {
 
     gid_tab = pr_table_alloc(auth_pool, 0);
 
-    if (pr_table_ctl(gid_tab, PR_TABLE_CTL_SET_KEY_CMP, gid_keycmp_cb) < 0) {
+    if (pr_table_ctl(gid_tab, PR_TABLE_CTL_SET_KEY_CMP,
+        (void *) gid_keycmp_cb) < 0) {
       pr_trace_msg(trace_channel, 2,
         "error setting key comparison callback for gidcache: %s",
         strerror(errno));
       ok = FALSE;
     }
 
-    if (pr_table_ctl(gid_tab, PR_TABLE_CTL_SET_KEY_HASH, gid_hash_cb) < 0) {
+    if (pr_table_ctl(gid_tab, PR_TABLE_CTL_SET_KEY_HASH,
+        (void *) gid_hash_cb) < 0) {
       pr_trace_msg(trace_channel, 2,
         "error setting key hash callback for gidcache: %s",
         strerror(errno));
@@ -179,6 +185,10 @@ static void gidcache_create(void) {
 }
 
 static void gidcache_add(gid_t gid, const char *name) {
+  if (!(auth_caching & PR_AUTH_CACHE_FL_GID2NAME)) {
+    return;
+  }
+
   gidcache_create();
 
   if (gid_tab) {
@@ -257,6 +267,14 @@ static modret_t *dispatch_auth(cmd_rec *cmd, char *match, module **m) {
 
   start_tab = pr_stash_get_symbol(PR_SYM_AUTH, match, NULL,
     &cmd->stash_index);
+  if (start_tab == NULL) {
+    int xerrno = errno;
+
+    pr_trace_msg(trace_channel, 1, "error finding start symbol for '%s': %s",
+      match, strerror(xerrno));
+    return PR_ERROR_MSG(cmd, NULL, strerror(xerrno));
+  }
+
   iter_tab = start_tab;
 
   while (iter_tab) {
@@ -272,18 +290,29 @@ static modret_t *dispatch_auth(cmd_rec *cmd, char *match, module **m) {
 
     mr = pr_module_call(iter_tab->m, iter_tab->handler, cmd);
 
-    if (iter_tab->auth_flags & PR_AUTH_FL_REQUIRED)
+    /* Return a pointer, if requested, to the module which answered the
+     * auth request.  This is used, for example, by auth_getpwnam() for
+     * associating the answering auth module with the data looked up.
+     */
+
+    if (iter_tab->auth_flags & PR_AUTH_FL_REQUIRED) {
+      pr_trace_msg(trace_channel, 6,
+        "\"%s\" response from module mod_%s is authoritative", match,
+        iter_tab->m->name);
+
+      if (m) {
+        *m = iter_tab->m;
+      }
+
       break;
+    }
 
     if (MODRET_ISHANDLED(mr) ||
         MODRET_ISERROR(mr)) {
 
-      /* Return a pointer, if requested, to the module which answered the
-       * auth request.  This is used, for example, by auth_getpwnam() for
-       * associating the answering auth module with the data looked up.
-       */
-      if (m)
+      if (m) {
         *m = iter_tab->m;
+      }
 
       break;
     }
@@ -296,6 +325,7 @@ static modret_t *dispatch_auth(cmd_rec *cmd, char *match, module **m) {
       /* We have looped back to the start.  Break out now and do not loop
        * around again (and again, and again...)
        */
+      pr_trace_msg(trace_channel, 15, "reached end of symbols for '%s'", match);
       mr = PR_DECLINED(cmd);
       break;
     }
@@ -306,10 +336,9 @@ static modret_t *dispatch_auth(cmd_rec *cmd, char *match, module **m) {
 
 void pr_auth_setpwent(pool *p) {
   cmd_rec *cmd = NULL;
-  modret_t *mr = NULL;
 
   cmd = make_cmd(p, 0);
-  mr = dispatch_auth(cmd, "setpwent", NULL);
+  (void) dispatch_auth(cmd, "setpwent", NULL);
 
   if (cmd->tmp_pool) {
     destroy_pool(cmd->tmp_pool);
@@ -321,10 +350,9 @@ void pr_auth_setpwent(pool *p) {
 
 void pr_auth_endpwent(pool *p) {
   cmd_rec *cmd = NULL;
-  modret_t *mr = NULL;
 
   cmd = make_cmd(p, 0);
-  mr = dispatch_auth(cmd, "endpwent", NULL);
+  (void) dispatch_auth(cmd, "endpwent", NULL);
 
   if (cmd->tmp_pool) {
     destroy_pool(cmd->tmp_pool);
@@ -343,10 +371,9 @@ void pr_auth_endpwent(pool *p) {
 
 void pr_auth_setgrent(pool *p) {
   cmd_rec *cmd = NULL;
-  modret_t *mr = NULL;
 
   cmd = make_cmd(p, 0);
-  mr = dispatch_auth(cmd, "setgrent", NULL);
+  (void) dispatch_auth(cmd, "setgrent", NULL);
 
   if (cmd->tmp_pool) {
     destroy_pool(cmd->tmp_pool);
@@ -358,10 +385,9 @@ void pr_auth_setgrent(pool *p) {
 
 void pr_auth_endgrent(pool *p) {
   cmd_rec *cmd = NULL;
-  modret_t *mr = NULL;
 
   cmd = make_cmd(p, 0);
-  mr = dispatch_auth(cmd, "endgrent", NULL);
+  (void) dispatch_auth(cmd, "endgrent", NULL);
 
   if (cmd->tmp_pool) {
     destroy_pool(cmd->tmp_pool);
@@ -379,8 +405,10 @@ struct passwd *pr_auth_getpwent(pool *p) {
   cmd = make_cmd(p, 0);
   mr = dispatch_auth(cmd, "getpwent", NULL);
 
-  if (MODRET_ISHANDLED(mr) && MODRET_HASDATA(mr))
+  if (MODRET_ISHANDLED(mr) &&
+      MODRET_HASDATA(mr)) {
     res = mr->data;
+  }
 
   if (cmd->tmp_pool) {
     destroy_pool(cmd->tmp_pool);
@@ -393,12 +421,12 @@ struct passwd *pr_auth_getpwent(pool *p) {
 
   /* Make sure the UID and GID are not -1 */
   if (res->pw_uid == (uid_t) -1) {
-    pr_log_pri(PR_LOG_ERR, "error: UID of -1 not allowed");
+    pr_log_pri(PR_LOG_WARNING, "error: UID of -1 not allowed");
     return NULL;
   }
 
   if (res->pw_gid == (gid_t) -1) {
-    pr_log_pri(PR_LOG_ERR, "error: GID of -1 not allowed");
+    pr_log_pri(PR_LOG_WARNING, "error: GID of -1 not allowed");
     return NULL;
   }
 
@@ -413,8 +441,10 @@ struct group *pr_auth_getgrent(pool *p) {
   cmd = make_cmd(p, 0);
   mr = dispatch_auth(cmd, "getgrent", NULL);
 
-  if (MODRET_ISHANDLED(mr) && MODRET_HASDATA(mr))
+  if (MODRET_ISHANDLED(mr) &&
+      MODRET_HASDATA(mr)) {
     res = mr->data;
+  }
 
   if (cmd->tmp_pool) {
     destroy_pool(cmd->tmp_pool);
@@ -427,7 +457,7 @@ struct group *pr_auth_getgrent(pool *p) {
 
   /* Make sure the GID is not -1 */
   if (res->gr_gid == (gid_t) -1) {
-    pr_log_pri(PR_LOG_ERR, "error: GID of -1 not allowed");
+    pr_log_pri(PR_LOG_WARNING, "error: GID of -1 not allowed");
     return NULL;
   }
 
@@ -444,8 +474,9 @@ struct passwd *pr_auth_getpwnam(pool *p, const char *name) {
   mr = dispatch_auth(cmd, "getpwnam", &m);
 
   if (MODRET_ISHANDLED(mr) &&
-      MODRET_HASDATA(mr))
+      MODRET_HASDATA(mr)) {
     res = mr->data;
+  }
 
   if (cmd->tmp_pool) {
     destroy_pool(cmd->tmp_pool);
@@ -460,12 +491,12 @@ struct passwd *pr_auth_getpwnam(pool *p, const char *name) {
 
   /* Make sure the UID and GID are not -1 */
   if (res->pw_uid == (uid_t) -1) {
-    pr_log_pri(PR_LOG_ERR, "error: UID of -1 not allowed");
+    pr_log_pri(PR_LOG_WARNING, "error: UID of -1 not allowed");
     return NULL;
   }
 
   if (res->pw_gid == (gid_t) -1) {
-    pr_log_pri(PR_LOG_ERR, "error: GID of -1 not allowed");
+    pr_log_pri(PR_LOG_WARNING, "error: GID of -1 not allowed");
     return NULL;
   }
 
@@ -512,7 +543,7 @@ struct passwd *pr_auth_getpwnam(pool *p, const char *name) {
     }
   }
 
-  uidcache_add(res->pw_uid, name);
+  uidcache_add(res->pw_uid, res->pw_name);
 
   /* Get the (possibly rewritten) home directory. */
   res->pw_dir = pr_auth_get_home(p, res->pw_dir);
@@ -530,8 +561,10 @@ struct passwd *pr_auth_getpwuid(pool *p, uid_t uid) {
   cmd = make_cmd(p, 1, (void *) &uid);
   mr = dispatch_auth(cmd, "getpwuid", NULL);
 
-  if (MODRET_ISHANDLED(mr) && MODRET_HASDATA(mr))
+  if (MODRET_ISHANDLED(mr) &&
+      MODRET_HASDATA(mr)) {
     res = mr->data;
+  }
 
   if (cmd->tmp_pool) {
     destroy_pool(cmd->tmp_pool);
@@ -546,12 +579,12 @@ struct passwd *pr_auth_getpwuid(pool *p, uid_t uid) {
 
   /* Make sure the UID and GID are not -1 */
   if (res->pw_uid == (uid_t) -1) {
-    pr_log_pri(PR_LOG_ERR, "error: UID of -1 not allowed");
+    pr_log_pri(PR_LOG_WARNING, "error: UID of -1 not allowed");
     return NULL;
   }
 
   if (res->pw_gid == (gid_t) -1) {
-    pr_log_pri(PR_LOG_ERR, "error: GID of -1 not allowed");
+    pr_log_pri(PR_LOG_WARNING, "error: GID of -1 not allowed");
     return NULL;
   }
 
@@ -568,8 +601,10 @@ struct group *pr_auth_getgrnam(pool *p, const char *name) {
   cmd = make_cmd(p, 1, name);
   mr = dispatch_auth(cmd, "getgrnam", NULL);
 
-  if (MODRET_ISHANDLED(mr) && MODRET_HASDATA(mr))
+  if (MODRET_ISHANDLED(mr) &&
+      MODRET_HASDATA(mr)) {
     res = mr->data;
+  }
 
   if (cmd->tmp_pool) {
     destroy_pool(cmd->tmp_pool);
@@ -584,7 +619,7 @@ struct group *pr_auth_getgrnam(pool *p, const char *name) {
 
   /* Make sure the GID is not -1 */
   if (res->gr_gid == (gid_t) -1) {
-    pr_log_pri(PR_LOG_ERR, "error: GID of -1 not allowed");
+    pr_log_pri(PR_LOG_WARNING, "error: GID of -1 not allowed");
     return NULL;
   }
 
@@ -603,8 +638,10 @@ struct group *pr_auth_getgrgid(pool *p, gid_t gid) {
   cmd = make_cmd(p, 1, (void *) &gid);
   mr = dispatch_auth(cmd, "getgrgid", NULL);
 
-  if (MODRET_ISHANDLED(mr) && MODRET_HASDATA(mr))
+  if (MODRET_ISHANDLED(mr) &&
+      MODRET_HASDATA(mr)) {
     res = mr->data;
+  }
 
   if (cmd->tmp_pool) {
     destroy_pool(cmd->tmp_pool);
@@ -619,7 +656,7 @@ struct group *pr_auth_getgrgid(pool *p, gid_t gid) {
 
   /* Make sure the GID is not -1 */
   if (res->gr_gid == (gid_t) -1) {
-    pr_log_pri(PR_LOG_ERR, "error: GID of -1 not allowed");
+    pr_log_pri(PR_LOG_WARNING, "error: GID of -1 not allowed");
     return NULL;
   }
 
@@ -698,11 +735,12 @@ int pr_auth_authenticate(pool *p, const char *name, const char *pw) {
 
   mr = dispatch_auth(cmd, "auth", m ? &m : NULL);
 
-  if (MODRET_ISHANDLED(mr))
+  if (MODRET_ISHANDLED(mr)) {
     res = MODRET_HASDATA(mr) ? PR_AUTH_RFC2228_OK : PR_AUTH_OK;
 
-  else if (MODRET_ISERROR(mr))
+  } else if (MODRET_ISERROR(mr)) {
     res = MODRET_ERROR(mr);
+  }
 
   if (cmd->tmp_pool) {
     destroy_pool(cmd->tmp_pool);
@@ -821,8 +859,9 @@ int pr_auth_check(pool *p, const char *cpw, const char *name, const char *pw) {
 
   mr = dispatch_auth(cmd, "check", m ? &m : NULL);
 
-  if (MODRET_ISHANDLED(mr))
+  if (MODRET_ISHANDLED(mr)) {
     res = MODRET_HASDATA(mr) ? PR_AUTH_RFC2228_OK : PR_AUTH_OK;
+  }
 
   if (cmd->tmp_pool) {
     destroy_pool(cmd->tmp_pool);
@@ -840,11 +879,12 @@ int pr_auth_requires_pass(pool *p, const char *name) {
   cmd = make_cmd(p, 1, name);
   mr = dispatch_auth(cmd, "requires_pass", NULL);
 
-  if (MODRET_ISHANDLED(mr))
+  if (MODRET_ISHANDLED(mr)) {
     res = FALSE;
 
-  else if (MODRET_ISERROR(mr))
+  } else if (MODRET_ISERROR(mr)) {
     res = MODRET_ERROR(mr);
+  }
 
   if (cmd->tmp_pool) {
     destroy_pool(cmd->tmp_pool);
@@ -865,7 +905,8 @@ const char *pr_auth_uid2name(pool *p, uid_t uid) {
 
   uidcache_create();
 
-  if (uid_tab) {
+  if ((auth_caching & PR_AUTH_CACHE_FL_UID2NAME) &&
+      uid_tab) {
     void *v = NULL;
 
     v = pr_table_kget(uid_tab, (const void *) &uid, sizeof(uid_t), NULL);
@@ -923,7 +964,8 @@ const char *pr_auth_gid2name(pool *p, gid_t gid) {
 
   gidcache_create();
 
-  if (gid_tab) {
+  if ((auth_caching & PR_AUTH_CACHE_FL_GID2NAME) &&
+       gid_tab) {
     void *v = NULL;
  
     v = pr_table_kget(gid_tab, (const void *) &gid, sizeof(gid_t), NULL);
@@ -978,10 +1020,12 @@ uid_t pr_auth_name2uid(pool *p, const char *name) {
   cmd = make_cmd(p, 1, name);
   mr = dispatch_auth(cmd, "name2uid", NULL);
 
-  if (MODRET_ISHANDLED(mr))
+  if (MODRET_ISHANDLED(mr)) {
     res = *((uid_t *) mr->data);
-  else
+
+  } else {
     errno = EINVAL;
+  }
 
   if (cmd->tmp_pool) {
     destroy_pool(cmd->tmp_pool);
@@ -999,10 +1043,12 @@ gid_t pr_auth_name2gid(pool *p, const char *name) {
   cmd = make_cmd(p, 1, name);
   mr = dispatch_auth(cmd, "name2gid", NULL);
 
-  if (MODRET_ISHANDLED(mr))
+  if (MODRET_ISHANDLED(mr)) {
     res = *((gid_t *) mr->data);
-  else
+
+  } else {
     errno = EINVAL;
+  }
 
   if (cmd->tmp_pool) {
     destroy_pool(cmd->tmp_pool);
@@ -1031,7 +1077,8 @@ int pr_auth_getgroups(pool *p, const char *name, array_header **group_ids,
 
   mr = dispatch_auth(cmd, "getgroups", NULL);
 
-  if (MODRET_ISHANDLED(mr) && MODRET_HASDATA(mr)) {
+  if (MODRET_ISHANDLED(mr) &&
+      MODRET_HASDATA(mr)) {
     res = *((int *) mr->data);
 
     /* Note: the number of groups returned should, barring error,
@@ -1051,11 +1098,13 @@ int pr_auth_getgroups(pool *p, const char *name, array_header **group_ids,
         snprintf(buf, sizeof(buf)-1, "%lu", (unsigned long) gids[i]);
         buf[sizeof(buf)-1] = '\0';
 
+        pr_signals_handle();
         strgids = pstrcat(p, strgids, i != 0 ? ", " : "", buf, NULL);
       }
 
       pr_log_debug(DEBUG10, "retrieved group %s: %s",
-        (*group_ids)->nelts == 1 ? "ID" : "IDs", strgids);
+        (*group_ids)->nelts == 1 ? "ID" : "IDs",
+        *strgids ? strgids : "(None; corrupted group file?)");
     }
 
     if (group_names) {
@@ -1063,11 +1112,14 @@ int pr_auth_getgroups(pool *p, const char *name, array_header **group_ids,
       char *strgroups = ""; 
       char **groups = (*group_names)->elts;
 
-      for (i = 0; i < (*group_names)->nelts; i++)
+      for (i = 0; i < (*group_names)->nelts; i++) {
+        pr_signals_handle();
         strgroups = pstrcat(p, strgroups, i != 0 ? ", " : "", groups[i], NULL);
-    
+      }
+ 
       pr_log_debug(DEBUG10, "retrieved group %s: %s",
-        (*group_names)->nelts == 1 ? "name" : "names", strgroups);
+        (*group_names)->nelts == 1 ? "name" : "names",
+        *strgroups ? strgroups : "(None; corrupted group file?)");
     }
   }
 
@@ -1080,11 +1132,12 @@ int pr_auth_getgroups(pool *p, const char *name, array_header **group_ids,
 }
 
 /* This is one messy function.  Yuck.  Yay legacy code. */
-config_rec *pr_auth_get_anon_config(pool *p, char **login_name,
-    char **user_name, char **anon_name) {
-  config_rec *c = NULL, *topc = NULL, *anon_c = NULL;
-  char *config_user_name, *config_anon_name = NULL;
+config_rec *pr_auth_get_anon_config(pool *p, char **login_user,
+    char **real_user, char **anon_name) {
+  config_rec *c = NULL, *alias_config = NULL, *anon_config = NULL;
+  char *config_user_name = NULL, *config_anon_name = NULL;
   unsigned char is_alias = FALSE, *auth_alias_only = NULL;
+  unsigned long config_flags = (PR_CONFIG_FIND_FL_SKIP_DIR|PR_CONFIG_FIND_FL_SKIP_LIMIT|PR_CONFIG_FIND_FL_SKIP_DYNDIR);
 
   /* Precendence rules:
    *   1. Search for UserAlias directive.
@@ -1093,9 +1146,9 @@ config_rec *pr_auth_get_anon_config(pool *p, char **login_name,
    */
 
   config_user_name = get_param_ptr(main_server->conf, "UserName", FALSE);
-  if (config_user_name &&
-      user_name) {
-    *user_name = config_user_name;
+  if (config_user_name != NULL &&
+      real_user != NULL) {
+    *real_user = config_user_name;
   }
 
   /* If the main_server->conf->set list is large (e.g. there are many
@@ -1110,63 +1163,78 @@ config_rec *pr_auth_get_anon_config(pool *p, char **login_name,
    * config contexts it might be.
    */
 
-  c = find_config(main_server->conf, CONF_PARAM, "UserAlias", TRUE);
-  if (c) {
+  c = find_config2(main_server->conf, CONF_PARAM, "UserAlias", TRUE,
+    config_flags);
+  if (c != NULL) {
     do {
+      const char *alias;
+
       pr_signals_handle();
 
-      if (strcmp(c->argv[0], "*") == 0 ||
-          strcmp(c->argv[0], *login_name) == 0) {
+      alias = c->argv[0];
+      if (strncmp(alias, "*", 2) == 0 ||
+          strcmp(alias, *login_user) == 0) {
         is_alias = TRUE;
+        alias_config = c;
         break;
       }
 
-    } while ((c = find_config_next(c, c->next, CONF_PARAM, "UserAlias",
-      TRUE)) != NULL);
+    } while ((c = find_config_next2(c, c->next, CONF_PARAM, "UserAlias",
+      TRUE, config_flags)) != NULL);
   }
 
   /* This is where things get messy, rapidly. */
-  topc = c;
+  if (is_alias == TRUE) {
+    c = alias_config;
+  }
 
   while (c && c->parent &&
-    (auth_alias_only = get_param_ptr(c->parent->set, "AuthAliasOnly", FALSE))) {
+    (auth_alias_only = get_param_ptr(c->parent->subset, "AuthAliasOnly", FALSE))) {
 
     /* while() loops should always handle signals. */
     pr_signals_handle();
 
-    if (auth_alias_only) {
-      /* If AuthAliasOnly is on, ignore this one and continue. */
-      if (*auth_alias_only == TRUE) {
-        c = find_config_next(c, c->next, CONF_PARAM, "UserAlias", TRUE);
-        continue;
-      }
+    /* If AuthAliasOnly is on, ignore this one and continue. */
+    if (auth_alias_only != NULL &&
+        *auth_alias_only == TRUE) {
+      c = find_config_next2(c, c->next, CONF_PARAM, "UserAlias", TRUE,
+        config_flags);
+      continue;
     }
 
     /* At this point, we have found an "AuthAliasOnly off" config in
-     * c->parent->set.  See if there's a UserAlias in the same config set.
+     * c->parent->set (which means that we cannot use the UserAlias, and thus
+     * is_alias is set to false).  See if there's a UserAlias in the same
+     * config set.
      */
 
     is_alias = FALSE;
 
-    find_config_set_top(topc);
-    c = find_config_next(c, c->next, CONF_PARAM, "UserAlias", TRUE);
+    find_config_set_top(alias_config);
+    c = find_config_next2(c, c->next, CONF_PARAM, "UserAlias", TRUE,
+      config_flags);
 
     if (c &&
-        (strcmp(c->argv[0], "*") == 0 ||
-         strcmp(c->argv[0], *login_name) == 0)) {
+        (strncmp(c->argv[0], "*", 2) == 0 ||
+         strcmp(c->argv[0], *login_user) == 0)) {
       is_alias = TRUE;
+      alias_config = c;
     }
   }
 
-  if (c) {
-    *login_name = c->argv[1];
+  /* At this point in time, c is guaranteed (if not null) to be pointing at
+   * a UserAlias config, either the original OR one found in the AuthAliasOnly
+   * config set.
+   */
+  if (c != NULL) {
+    *login_user = c->argv[1];
 
     /* If the alias is applied inside an <Anonymous> context, we have found
-     * our anon block.
+     * our <Anonymous> section.
      */
     if (c->parent &&
         c->parent->config_type == CONF_ANON) {
-      c = c->parent;
+      anon_config = c->parent;
 
     } else {
       c = NULL;
@@ -1174,68 +1242,106 @@ config_rec *pr_auth_get_anon_config(pool *p, char **login_name,
   }
 
   /* Next, search for an anonymous entry. */
-
-  if (c == NULL) {
+  if (anon_config == NULL) {
     c = find_config(main_server->conf, CONF_ANON, NULL, FALSE);
 
   } else {
-    find_config_set_top(c);
-    anon_c = c;
+    find_config_set_top(anon_config);
+    c = anon_config;
   }
 
-  if (c) {
+  /* If anon_config is null here but c is not null, then we may have found
+   * a candidate <Anonymous> section.  Let's examine it more closely.
+   */
+  if (c != NULL) {
+    config_rec *starting_c;
+
+    starting_c = c;
     do {
       pr_signals_handle();
 
       config_anon_name = get_param_ptr(c->subset, "UserName", FALSE);
-
-      if (!config_anon_name)
+      if (config_anon_name == NULL) {
         config_anon_name = config_user_name;
+      }
 
-      if (config_anon_name &&
-          strcmp(config_anon_name, *login_name) == 0) {
-         if (anon_name)
-           *anon_name = config_anon_name;
-         break;
+      if (config_anon_name != NULL &&
+          strcmp(config_anon_name, *login_user) == 0) {
+
+        /* We found our <Anonymous> section. */
+        anon_config = c;
+
+        if (anon_name != NULL) {
+          *anon_name = config_anon_name;
+        }
+        break;
       }
  
     } while ((c = find_config_next(c, c->next, CONF_ANON, NULL,
       FALSE)) != NULL);
+
+    c = starting_c;
   }
 
-  if (!is_alias) {
-    /* Yes, we do want to be using c, not anon_c, here.  Otherwise, we
-     * risk a regression of Bug#3501.
-     */
-
+  if (is_alias == FALSE) {
     auth_alias_only = get_param_ptr(c ? c->subset : main_server->conf,
       "AuthAliasOnly", FALSE);
 
-    if (auth_alias_only &&
+    if (auth_alias_only != NULL &&
         *auth_alias_only == TRUE) {
-      if (c &&
+      if (c != NULL &&
           c->config_type == CONF_ANON) {
         c = NULL;
 
       } else {
-        *login_name = NULL;
+        *login_user = NULL;
       }
 
-      auth_alias_only = get_param_ptr(main_server->conf, "AuthAliasOnly",
-        FALSE);
-      if (*login_name &&
-          auth_alias_only &&
-          *auth_alias_only == TRUE)
-        *login_name = NULL;
+      /* Note: We only need to look for AuthAliasOnly in main_server IFF
+       * c is NOT null.  If c IS null, then we will already have looked up
+       * AuthAliasOnly in main_server above.
+       */
+      if (c != NULL) {
+        auth_alias_only = get_param_ptr(main_server->conf, "AuthAliasOnly",
+          FALSE);
+      }
 
-      if ((!login_name || !c) &&
-          anon_name) {
+      if (login_user != NULL &&
+          auth_alias_only != NULL &&
+          *auth_alias_only == TRUE) {
+        *login_user = NULL;
+      }
+
+      if ((login_user == NULL || anon_config == NULL) &&
+          anon_name != NULL) {
         *anon_name = NULL;
       }
     }
+
+  } else {
+    config_rec *alias_parent_config = NULL;
+
+    /* We have found a matching UserAlias for the USER name sent by the client.
+     * But we need to properly handle any AuthAliasOnly directives in that
+     * config as well (Bug#2070).
+     */
+    if (alias_config != NULL) {
+      alias_parent_config = alias_config->parent;
+    }
+
+    auth_alias_only = get_param_ptr(alias_parent_config ?
+      alias_parent_config->subset : main_server->conf, "AuthAliasOnly", FALSE);
+
+    if (auth_alias_only != NULL &&
+        *auth_alias_only == TRUE) {
+      if (alias_parent_config != NULL &&
+          alias_parent_config->config_type == CONF_ANON) {
+        anon_config = alias_parent_config;
+      }
+    }
   }
 
-  return c;
+  return anon_config;
 }
 
 int pr_auth_banned_by_ftpusers(xaset_t *ctx, const char *user) {
@@ -1267,7 +1373,7 @@ int pr_auth_banned_by_ftpusers(xaset_t *ctx, const char *user) {
       CHOP(buf);
 
       ptr = buf;
-      while (isspace((int) *ptr) && *ptr) {
+      while (PR_ISSPACE(*ptr) && *ptr) {
         ptr++;
       }
 
@@ -1332,7 +1438,8 @@ int pr_auth_is_valid_shell(xaset_t *ctx, const char *shell) {
 }
 
 int pr_auth_chroot(const char *path) {
-  int res;
+  int res, xerrno = 0;
+  time_t now;
 
 #if defined(HAVE_SETENV) && defined(__GLIBC__) && defined(__GLIBC_MINOR__) && \
   __GLIBC__ == 2 && __GLIBC_MINOR__ >= 3
@@ -1351,15 +1458,27 @@ int pr_auth_chroot(const char *path) {
   }
 #endif
 
-  pr_log_pri(PR_LOG_INFO, "Preparing to chroot to directory '%s'", path);
+  pr_log_debug(DEBUG1, "Preparing to chroot to directory '%s'", path);
+
+  /* Prepare for chroots and the ensuing timezone chicanery by calling
+   * our pr_localtime() routine now, which will cause libc (via localtime(2))
+   * to load the tzinfo data into memory, and hopefully retain it (Bug#3431).
+   */
+  now = time(NULL);
+  (void) pr_localtime(NULL, &now);
+
+  pr_event_generate("core.chroot", path);
 
   PRIVS_ROOT
   res = pr_fsio_chroot(path);
+  xerrno = errno;
   PRIVS_RELINQUISH
 
   if (res < 0) {
     pr_log_pri(PR_LOG_ERR, "chroot to '%s' failed for user '%s': %s", path,
-      session.user, strerror(errno));
+      session.user, strerror(xerrno));
+
+    errno = xerrno;
     return -1;
   }
 
@@ -1378,28 +1497,43 @@ int set_groups(pool *p, gid_t primary_gid, array_header *suppl_gids) {
   char *strgids = "";
 
   /* sanity check */
-  if (!p || !suppl_gids)
-    return 0;
+  if (p == NULL ||
+      suppl_gids == NULL) {
+
+# ifndef PR_DEVEL_COREDUMP
+    /* Set the primary GID of the process. */
+    res = setgid(primary_gid);
+# endif /* PR_DEVEL_COREDUMP */
+
+    return res;
+  }
+
+  ngids = suppl_gids->nelts;
+  gids = suppl_gids->elts;
+
+  if (ngids == 0 ||
+      gids == NULL) {
+    /* No supplemental GIDs to process. */
+
+# ifndef PR_DEVEL_COREDUMP
+    /* Set the primary GID of the process. */
+    res = setgid(primary_gid);
+# endif /* PR_DEVEL_COREDUMP */
+
+    return res;
+  }
 
   tmp_pool = make_sub_pool(p);
   pr_pool_tag(tmp_pool, "set_groups() tmp pool");
 
-  /* Check for a NULL supplemental group ID list. */
-  if (suppl_gids) {
-    ngids = suppl_gids->nelts;
-    gids = suppl_gids->elts;
+  proc_gids = pcalloc(tmp_pool, sizeof(gid_t) * (ngids));
 
-    if (ngids && gids) {
-      proc_gids = pcalloc(tmp_pool, sizeof(gid_t) * (ngids));
-
-      /* Note: the list of supplemental GIDs may contain duplicates.  Sort
-       * through the list and keep only the unique IDs - this should help avoid
-       * running into the NGROUPS limit when possible.  This algorithm may slow
-       * things down some; optimize it if/when possible.
-       */
-      proc_gids[nproc_gids++] = gids[0];
-    }
-  }
+  /* Note: the list of supplemental GIDs may contain duplicates.  Sort
+   * through the list and keep only the unique IDs - this should help avoid
+   * running into the NGROUPS limit when possible.  This algorithm may slow
+   * things down some; optimize it if/when possible.
+   */
+  proc_gids[nproc_gids++] = gids[0];
 
   for (i = 1; i < ngids; i++) {
     register unsigned int j = 0;
