@@ -1,6 +1,6 @@
 /*
  * ProFTPD - FTP server daemon
- * Copyright (c) 2001-2010 The ProFTPD Project team
+ * Copyright (c) 2001-2014 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -14,7 +14,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
  *
  * As a special exemption, the ProFTPD Project team and other respective
  * copyright holders give permission to link this program with OpenSSL, and
@@ -23,11 +23,10 @@
  */
 
 /* NetIO routines
- * $Id: netio.c,v 1.39.2.2 2011/02/26 02:46:45 castaglia Exp $
+ * $Id: netio.c,v 1.63 2014-01-06 06:57:16 castaglia Exp $
  */
 
 #include "conf.h"
-#include <signal.h>
 
 /* See RFC 854 for the definition of these Telnet values */
 
@@ -64,9 +63,9 @@
 
 static const char *trace_channel = "netio";
 
-static pr_netio_t *core_ctrl_netio = NULL, *ctrl_netio = NULL;
-static pr_netio_t *core_data_netio = NULL, *data_netio = NULL;
-static pr_netio_t *core_othr_netio = NULL, *othr_netio = NULL;
+static pr_netio_t *default_ctrl_netio = NULL, *ctrl_netio = NULL;
+static pr_netio_t *default_data_netio = NULL, *data_netio = NULL;
+static pr_netio_t *default_othr_netio = NULL, *othr_netio = NULL;
 
 /* Used to track whether the previous text read from the client's control
  * connection was a properly-terminated command.  If so, then read in the
@@ -99,13 +98,23 @@ static pr_netio_stream_t *netio_stream_alloc(pool *parent_pool) {
   nstrm->strm_data = NULL;
   nstrm->strm_errno = 0;
 
+  /* This table will not contain that many entries, so a low number
+   * of chains should suffice.
+   */
+  nstrm->notes = pr_table_nalloc(nstrm->strm_pool, 0, 4);
+
   return nstrm;
 }
 
-static pr_buffer_t *netio_buffer_alloc(pr_netio_stream_t *nstrm) {
+pr_buffer_t *pr_netio_buffer_alloc(pr_netio_stream_t *nstrm) {
   size_t bufsz;
   pr_buffer_t *pbuf = NULL;
 
+  if (nstrm == NULL) {
+    errno = EINVAL;
+    return NULL;
+  }
+
   pbuf = pcalloc(nstrm->strm_pool, sizeof(pr_buffer_t));
 
   /* Allocate a buffer. */
@@ -181,6 +190,20 @@ static int core_netio_poll_cb(pr_netio_stream_t *nstrm) {
   tval.tv_usec = 0;
 
   res = select(nstrm->strm_fd + 1, rfdsp, wfdsp, NULL, &tval);
+  while (res < 0) {
+    int xerrno = errno;
+
+    /* Watch for EAGAIN, and handle it by delaying temporarily. */
+    if (xerrno == EAGAIN) {
+      errno = EINTR;
+      pr_signals_handle();
+      continue;
+    }
+
+    errno = nstrm->strm_errno = xerrno;
+    break; 
+  }
+
   return res;
 }
 
@@ -196,8 +219,9 @@ static int core_netio_read_cb(pr_netio_stream_t *nstrm, char *buf,
 static pr_netio_stream_t *core_netio_reopen_cb(pr_netio_stream_t *nstrm, int fd,
     int mode) {
 
-  if (nstrm->strm_fd != -1)
+  if (nstrm->strm_fd != -1) {
     close(nstrm->strm_fd);
+  }
 
   nstrm->strm_fd = fd;
   nstrm->strm_mode = mode;
@@ -219,22 +243,31 @@ static int core_netio_write_cb(pr_netio_stream_t *nstrm, char *buf,
 
 void pr_netio_abort(pr_netio_stream_t *nstrm) {
 
-  if (!nstrm) {
+  if (nstrm == NULL) {
     errno = EINVAL;
     return;
   }
 
-  if (nstrm->strm_type == PR_NETIO_STRM_CTRL)
-    ctrl_netio ? ctrl_netio->abort(nstrm) :
-      core_ctrl_netio->abort(nstrm);
+  switch (nstrm->strm_type) {
+    case PR_NETIO_STRM_CTRL:
+      ctrl_netio ? (ctrl_netio->abort)(nstrm) :
+        (default_ctrl_netio->abort)(nstrm);
+      break;
 
-  if (nstrm->strm_type == PR_NETIO_STRM_DATA)
-    data_netio ? data_netio->abort(nstrm) :
-      core_data_netio->abort(nstrm);
+    case PR_NETIO_STRM_DATA:
+      data_netio ? (data_netio->abort)(nstrm) :
+        (default_data_netio->abort)(nstrm);
+      break;
 
-  if (nstrm->strm_type == PR_NETIO_STRM_OTHR)
-    othr_netio ? othr_netio->abort(nstrm) :
-      core_othr_netio->abort(nstrm);
+    case PR_NETIO_STRM_OTHR:
+      othr_netio ? (othr_netio->abort)(nstrm) :
+        (default_othr_netio->abort)(nstrm);
+      break;
+
+    default:
+      errno = EINVAL;
+      return;
+  }
 
   return;
 }
@@ -242,30 +275,29 @@ void pr_netio_abort(pr_netio_stream_t *nstrm) {
 int pr_netio_close(pr_netio_stream_t *nstrm) {
   int res = -1;
 
-  if (!nstrm) {
+  if (nstrm == NULL) {
     errno = EINVAL;
     return -1;
   }
 
-  if (nstrm->strm_type == PR_NETIO_STRM_CTRL) {
-    res = ctrl_netio ? ctrl_netio->close(nstrm) :
-      core_ctrl_netio->close(nstrm);
-    destroy_pool(nstrm->strm_pool);
-    return res;
-  }
+  switch (nstrm->strm_type) {
+    case PR_NETIO_STRM_CTRL:
+      res = ctrl_netio ? (ctrl_netio->close)(nstrm) :
+        (default_ctrl_netio->close)(nstrm);
+      destroy_pool(nstrm->strm_pool);
+      return res;
 
-  if (nstrm->strm_type == PR_NETIO_STRM_DATA) {
-    res = data_netio ? data_netio->close(nstrm) :
-      core_data_netio->close(nstrm);
-    destroy_pool(nstrm->strm_pool);
-    return res;
-  }
+    case PR_NETIO_STRM_DATA:
+      res = data_netio ? (data_netio->close)(nstrm) :
+        (default_data_netio->close)(nstrm);
+      destroy_pool(nstrm->strm_pool);
+      return res;
 
-  if (nstrm->strm_type == PR_NETIO_STRM_OTHR) {
-    res = othr_netio ? othr_netio->close(nstrm) :
-      core_othr_netio->close(nstrm);
-    destroy_pool(nstrm->strm_pool);
-    return res;
+    case PR_NETIO_STRM_OTHR:
+      res = othr_netio ? (othr_netio->close)(nstrm) :
+        (default_othr_netio->close)(nstrm);
+      destroy_pool(nstrm->strm_pool);
+      return res;
   }
 
   errno = EPERM;
@@ -276,17 +308,19 @@ static int netio_lingering_close(pr_netio_stream_t *nstrm, long linger,
     int flags) {
   int res;
 
-  if (!nstrm) {
+  if (nstrm == NULL) {
     errno = EINVAL;
     return -1;
   }
 
-  if (nstrm->strm_fd < 0)
+  if (nstrm->strm_fd < 0) {
     /* Already closed. */
     return 0;
+  }
 
-  if (!(flags & NETIO_LINGERING_CLOSE_FL_NO_SHUTDOWN))
+  if (!(flags & NETIO_LINGERING_CLOSE_FL_NO_SHUTDOWN)) {
     pr_netio_shutdown(nstrm, 1);
+  }
 
   if (nstrm->strm_fd >= 0) {
     struct timeval tv;
@@ -345,17 +379,19 @@ static int netio_lingering_close(pr_netio_stream_t *nstrm, long linger,
   pr_trace_msg(trace_channel, 8, "done lingering, closing fd %d",
     nstrm->strm_fd);
 
-  if (nstrm->strm_type == PR_NETIO_STRM_CTRL)
-    return ctrl_netio ? ctrl_netio->close(nstrm) :
-      core_ctrl_netio->close(nstrm);
+  switch (nstrm->strm_type) {
+    case PR_NETIO_STRM_CTRL:
+      return ctrl_netio ? (ctrl_netio->close)(nstrm) :
+        (default_ctrl_netio->close)(nstrm);
 
-  if (nstrm->strm_type == PR_NETIO_STRM_DATA)
-    return data_netio ? data_netio->close(nstrm) :
-      core_data_netio->close(nstrm);
+    case PR_NETIO_STRM_DATA:
+      return data_netio ? (data_netio->close)(nstrm) :
+        (default_data_netio->close)(nstrm);
 
-  if (nstrm->strm_type == PR_NETIO_STRM_OTHR)
-    return othr_netio ? othr_netio->close(nstrm) :
-      core_othr_netio->close(nstrm);
+    case PR_NETIO_STRM_OTHR:
+      return othr_netio ? (othr_netio->close)(nstrm) :
+        (default_othr_netio->close)(nstrm);
+  }
 
   errno = EPERM;
   return -1;
@@ -364,7 +400,7 @@ static int netio_lingering_close(pr_netio_stream_t *nstrm, long linger,
 int pr_netio_lingering_abort(pr_netio_stream_t *nstrm, long linger) {
   int res;
 
-  if (!nstrm) {
+  if (nstrm == NULL) {
     errno = EINVAL;
     return -1;
   }
@@ -421,7 +457,7 @@ pr_netio_stream_t *pr_netio_open(pool *parent_pool, int strm_type, int fd,
     int mode) {
   pr_netio_stream_t *nstrm = NULL;
 
-  if (!parent_pool) {
+  if (parent_pool == NULL) {
     errno = EINVAL;
     return NULL;
   }
@@ -429,25 +465,51 @@ pr_netio_stream_t *pr_netio_open(pool *parent_pool, int strm_type, int fd,
   /* Create a new stream object, then pass that the NetIO open handler. */
   nstrm = netio_stream_alloc(parent_pool);
 
-  if (strm_type == PR_NETIO_STRM_CTRL) {
-    nstrm->strm_type = PR_NETIO_STRM_CTRL;
-    nstrm->strm_mode = mode;
-    return ctrl_netio ? (ctrl_netio->open)(nstrm, fd, mode) :
-      (core_ctrl_netio->open)(nstrm, fd, mode);
-  }
+  switch (strm_type) {
+    case PR_NETIO_STRM_CTRL:
+      nstrm->strm_type = PR_NETIO_STRM_CTRL;
+      nstrm->strm_mode = mode;
 
-  if (strm_type == PR_NETIO_STRM_DATA) {
-    nstrm->strm_type = PR_NETIO_STRM_DATA;
-    nstrm->strm_mode = mode;
-    return data_netio ? (data_netio->open)(nstrm, fd, mode) :
-      (core_data_netio->open)(nstrm, fd, mode);
-  }
+      if (ctrl_netio != NULL) {
+        pr_table_add(nstrm->notes, pstrdup(nstrm->strm_pool, "core.netio"),
+          ctrl_netio, sizeof(pr_netio_t *));
+        return (ctrl_netio->open)(nstrm, fd, mode);
 
-  if (strm_type == PR_NETIO_STRM_OTHR) {
-    nstrm->strm_type = PR_NETIO_STRM_OTHR;
-    nstrm->strm_mode = mode;
-    return othr_netio ? (othr_netio->open)(nstrm, fd, mode) :
-      (core_othr_netio->open)(nstrm, fd, mode);
+      } else {
+        pr_table_add(nstrm->notes, pstrdup(nstrm->strm_pool, "core.netio"),
+          default_ctrl_netio, sizeof(pr_netio_t *));
+        return (default_ctrl_netio->open)(nstrm, fd, mode);
+      }
+
+    case PR_NETIO_STRM_DATA:
+      nstrm->strm_type = PR_NETIO_STRM_DATA;
+      nstrm->strm_mode = mode;
+
+      if (data_netio != NULL) {
+        pr_table_add(nstrm->notes, pstrdup(nstrm->strm_pool, "core.netio"),
+          data_netio, sizeof(pr_netio_t *));
+        return (data_netio->open)(nstrm, fd, mode);
+
+      } else {
+        pr_table_add(nstrm->notes, pstrdup(nstrm->strm_pool, "core.netio"),
+          default_data_netio, sizeof(pr_netio_t *));
+        return (default_data_netio->open)(nstrm, fd, mode);
+      }
+
+    case PR_NETIO_STRM_OTHR:
+      nstrm->strm_type = PR_NETIO_STRM_OTHR;
+      nstrm->strm_mode = mode;
+
+      if (othr_netio != NULL) {
+        pr_table_add(nstrm->notes, pstrdup(nstrm->strm_pool, "core.netio"),
+          othr_netio, sizeof(pr_netio_t *));
+        return (othr_netio->open)(nstrm, fd, mode);
+
+      } else {
+        pr_table_add(nstrm->notes, pstrdup(nstrm->strm_pool, "core.netio"),
+          default_othr_netio, sizeof(pr_netio_t *));
+        return (default_othr_netio->open)(nstrm, fd, mode);
+      }
   }
 
   destroy_pool(nstrm->strm_pool);
@@ -459,29 +521,31 @@ pr_netio_stream_t *pr_netio_open(pool *parent_pool, int strm_type, int fd,
 
 pr_netio_stream_t *pr_netio_reopen(pr_netio_stream_t *nstrm, int fd, int mode) {
 
-  if (!nstrm) {
+  if (nstrm == NULL) {
     errno = EINVAL;
     return NULL;
   }
 
-  if (nstrm->strm_type == PR_NETIO_STRM_CTRL)
-    return ctrl_netio ? (ctrl_netio->reopen)(nstrm, fd, mode) :
-      (core_ctrl_netio->reopen)(nstrm, fd, mode);
+  switch (nstrm->strm_type) {
+    case PR_NETIO_STRM_CTRL:
+      return ctrl_netio ? (ctrl_netio->reopen)(nstrm, fd, mode) :
+        (default_ctrl_netio->reopen)(nstrm, fd, mode);
 
-  if (nstrm->strm_type == PR_NETIO_STRM_DATA)
-    return data_netio ? (data_netio->reopen)(nstrm, fd, mode) :
-      (core_data_netio->reopen)(nstrm, fd, mode);
+    case PR_NETIO_STRM_DATA:
+      return data_netio ? (data_netio->reopen)(nstrm, fd, mode) :
+        (default_data_netio->reopen)(nstrm, fd, mode);
 
-  if (nstrm->strm_type == PR_NETIO_STRM_OTHR)
-    return othr_netio ? (othr_netio->reopen)(nstrm, fd, mode) :
-      (core_othr_netio->reopen)(nstrm, fd, mode);
+    case PR_NETIO_STRM_OTHR:
+      return othr_netio ? (othr_netio->reopen)(nstrm, fd, mode) :
+        (default_othr_netio->reopen)(nstrm, fd, mode);
+  }
 
   errno = EPERM;
   return NULL;
 }
 
 void pr_netio_reset_poll_interval(pr_netio_stream_t *nstrm) {
-  if (!nstrm) {
+  if (nstrm == NULL) {
     errno = EINVAL;
     return;
   }
@@ -492,7 +556,7 @@ void pr_netio_reset_poll_interval(pr_netio_stream_t *nstrm) {
 
 void pr_netio_set_poll_interval(pr_netio_stream_t *nstrm, unsigned int secs) {
 
-  if (!nstrm) {
+  if (nstrm == NULL) {
     errno = EINVAL;
     return;
   }
@@ -502,10 +566,10 @@ void pr_netio_set_poll_interval(pr_netio_stream_t *nstrm, unsigned int secs) {
 }
 
 int pr_netio_poll(pr_netio_stream_t *nstrm) {
-  int res = 0;
+  int res = 0, xerrno = 0;
 
   /* Sanity checks. */
-  if (!nstrm) {
+  if (nstrm == NULL) {
     errno = EINVAL;
     return -1;
   }
@@ -528,23 +592,24 @@ int pr_netio_poll(pr_netio_stream_t *nstrm) {
     switch (nstrm->strm_type) {
       case PR_NETIO_STRM_CTRL:
         res = ctrl_netio ? (ctrl_netio->poll)(nstrm) :
-          (core_ctrl_netio->poll)(nstrm);
+          (default_ctrl_netio->poll)(nstrm);
         break;
 
       case PR_NETIO_STRM_DATA:
         res = data_netio ? (data_netio->poll)(nstrm) :
-          (core_data_netio->poll)(nstrm);
+          (default_data_netio->poll)(nstrm);
         break;
 
       case PR_NETIO_STRM_OTHR:
         res = othr_netio ? (othr_netio->poll)(nstrm) :
-          (core_othr_netio->poll)(nstrm);
+          (default_othr_netio->poll)(nstrm);
         break;
     }
 
     switch (res) {
       case -1:
-        if (errno == EINTR) {
+        xerrno = errno;
+        if (xerrno == EINTR) {
           if (nstrm->strm_flags & PR_NETIO_SESS_ABORT) {
             nstrm->strm_flags &= ~PR_NETIO_SESS_ABORT;
             return 1;
@@ -556,13 +621,13 @@ int pr_netio_poll(pr_netio_stream_t *nstrm) {
         }
 
         /* Some other error occured */
-        nstrm->strm_errno = errno;
+        nstrm->strm_errno = xerrno;
 
         /* If this is the control stream, and the error indicates a
          * broken pipe (i.e. the client went away), AND there is a data
          * transfer is progress, abort the transfer.
          */
-        if (errno == EPIPE &&
+        if (xerrno == EPIPE &&
             nstrm->strm_type == PR_NETIO_STRM_CTRL &&
             (session.sf_flags & SF_XFER)) {
           pr_trace_msg(trace_channel, 5,
@@ -571,6 +636,7 @@ int pr_netio_poll(pr_netio_stream_t *nstrm) {
           session.sf_flags |= SF_ABORT;
         }
 
+        errno = nstrm->strm_errno;
         return -1;
 
       case 0:
@@ -602,42 +668,52 @@ int pr_netio_poll(pr_netio_stream_t *nstrm) {
 }
 
 int pr_netio_postopen(pr_netio_stream_t *nstrm) {
-  if (!nstrm) {
+  if (nstrm == NULL) {
     errno = EINVAL;
     return -1;
   }
 
-  if (nstrm->strm_type == PR_NETIO_STRM_CTRL)
-    return ctrl_netio ? (ctrl_netio->postopen)(nstrm) :
-      (core_ctrl_netio->postopen)(nstrm);
+  switch (nstrm->strm_type) {
+    case PR_NETIO_STRM_CTRL:
+      return ctrl_netio ? (ctrl_netio->postopen)(nstrm) :
+        (default_ctrl_netio->postopen)(nstrm);
 
-  if (nstrm->strm_type == PR_NETIO_STRM_DATA)
-    return data_netio ? (data_netio->postopen)(nstrm) :
-      (core_data_netio->postopen)(nstrm);
+    case PR_NETIO_STRM_DATA:
+      return data_netio ? (data_netio->postopen)(nstrm) :
+        (default_data_netio->postopen)(nstrm);
 
-  if (nstrm->strm_type == PR_NETIO_STRM_OTHR)
-    return othr_netio ? (othr_netio->postopen)(nstrm) :
-      (core_othr_netio->postopen)(nstrm);
+    case PR_NETIO_STRM_OTHR:
+      return othr_netio ? (othr_netio->postopen)(nstrm) :
+        (default_othr_netio->postopen)(nstrm);
+  }
 
   errno = EPERM;
   return -1;
 }
 
-int pr_netio_printf(pr_netio_stream_t *nstrm, const char *fmt, ...) {
-  va_list msg;
+int pr_netio_vprintf(pr_netio_stream_t *nstrm, const char *fmt, va_list msg) {
   char buf[PR_RESPONSE_BUFFER_SIZE] = {'\0'};
 
+  vsnprintf(buf, sizeof(buf), fmt, msg);
+  buf[sizeof(buf)-1] = '\0';
+
+  return pr_netio_write(nstrm, buf, strlen(buf));
+}
+
+int pr_netio_printf(pr_netio_stream_t *nstrm, const char *fmt, ...) {
+  int res;
+  va_list msg;
+
   if (!nstrm) {
     errno = EINVAL;
     return -1;
   }
 
   va_start(msg, fmt);
-  vsnprintf(buf, sizeof(buf), fmt, msg);
+  res = pr_netio_vprintf(nstrm, fmt, msg);
   va_end(msg);
-  buf[sizeof(buf)-1] = '\0';
 
-  return pr_netio_write(nstrm, buf, strlen(buf));
+  return res;
 }
 
 int pr_netio_printf_async(pr_netio_stream_t *nstrm, char *fmt, ...) {
@@ -659,6 +735,8 @@ int pr_netio_printf_async(pr_netio_stream_t *nstrm, char *fmt, ...) {
 
 int pr_netio_write(pr_netio_stream_t *nstrm, char *buf, size_t buflen) {
   int bwritten = 0, total = 0;
+  pr_buffer_t *pbuf;
+  pool *sub_pool;
 
   /* Sanity check */
   if (!nstrm) {
@@ -671,6 +749,43 @@ int pr_netio_write(pr_netio_stream_t *nstrm, char *buf, size_t buflen) {
     return -1;
   }
 
+  /* Before we send out the data to the client, generate an event
+   * for any listeners which may want to examine this data.  To do this, we
+   * need to allocate a pr_buffer_t for sending the buffer data to the
+   * listeners.
+   *
+   * We could just use nstrm->strm_pool, but for a long-lived control
+   * connection, this would amount to a slow memory increase.  So instead,
+   * we create a subpool from the stream's pool, and allocate the
+   * pr_buffer_t out of that.  Then simply destroy the subpool when done.
+   */
+
+  sub_pool = pr_pool_create_sz(nstrm->strm_pool, 64);
+  pbuf = pcalloc(sub_pool, sizeof(pr_buffer_t));
+  pbuf->buf = buf;
+  pbuf->buflen = buflen;
+  pbuf->current = pbuf->buf;
+  pbuf->remaining = 0;
+
+  switch (nstrm->strm_type) {
+    case PR_NETIO_STRM_CTRL:
+      pr_event_generate("core.ctrl-write", pbuf);
+      break;
+
+    case PR_NETIO_STRM_DATA:
+      pr_event_generate("core.data-write", pbuf);
+      break;
+
+    case PR_NETIO_STRM_OTHR:
+      pr_event_generate("core.othr-write", pbuf);
+      break;
+  }
+
+  /* The event listeners may have changed the data to write out. */
+  buf = pbuf->buf;
+  buflen = pbuf->buflen - pbuf->remaining;
+  destroy_pool(sub_pool);
+
   while (buflen) {
 
     switch (pr_netio_poll(nstrm)) {
@@ -689,7 +804,7 @@ int pr_netio_write(pr_netio_stream_t *nstrm, char *buf, size_t buflen) {
           switch (nstrm->strm_type) {
             case PR_NETIO_STRM_CTRL:
               bwritten = ctrl_netio ? (ctrl_netio->write)(nstrm, buf, buflen) :
-                (core_ctrl_netio->write)(nstrm, buf, buflen);
+                (default_ctrl_netio->write)(nstrm, buf, buflen);
                 break;
 
             case PR_NETIO_STRM_DATA:
@@ -697,12 +812,12 @@ int pr_netio_write(pr_netio_stream_t *nstrm, char *buf, size_t buflen) {
                 break;
 
               bwritten = data_netio ? (data_netio->write)(nstrm, buf, buflen) :
-                (core_data_netio->write)(nstrm, buf, buflen);
+                (default_data_netio->write)(nstrm, buf, buflen);
               break;
 
             case PR_NETIO_STRM_OTHR:
               bwritten = othr_netio ? (othr_netio->write)(nstrm, buf, buflen) :
-                (core_othr_netio->write)(nstrm, buf, buflen);
+                (default_othr_netio->write)(nstrm, buf, buflen);
               break;
           }
 
@@ -720,12 +835,14 @@ int pr_netio_write(pr_netio_stream_t *nstrm, char *buf, size_t buflen) {
     buflen -= bwritten;
   }
 
+  session.total_raw_out += total;
   return total;
 }
 
 int pr_netio_write_async(pr_netio_stream_t *nstrm, char *buf, size_t buflen) {
   int flags = 0;
   int bwritten = 0, total = 0;
+  pr_buffer_t *pbuf;
 
   /* Sanity check */
   if (!nstrm) {
@@ -745,6 +862,34 @@ int pr_netio_write_async(pr_netio_stream_t *nstrm, char *buf, size_t buflen) {
   if (fcntl(nstrm->strm_fd, F_SETFL, flags|O_NONBLOCK) == -1)
     return -1;
 
+  /* Before we send out the data to the client, generate an event
+   * for any listeners which may want to examine this data.
+   */
+
+  pbuf = pcalloc(nstrm->strm_pool, sizeof(pr_buffer_t));
+  pbuf->buf = buf;
+  pbuf->buflen = buflen;
+  pbuf->current = pbuf->buf;
+  pbuf->remaining = 0;
+
+  switch (nstrm->strm_type) {
+    case PR_NETIO_STRM_CTRL:
+      pr_event_generate("core.ctrl-write", pbuf);
+      break;
+
+    case PR_NETIO_STRM_DATA:
+      pr_event_generate("core.data-write", pbuf);
+      break;
+
+    case PR_NETIO_STRM_OTHR:
+      pr_event_generate("core.othr-write", pbuf);
+      break;
+  }
+
+  /* The event listeners may have changed the data to write out. */
+  buf = pbuf->buf;
+  buflen = pbuf->buflen - pbuf->remaining;
+
   while (buflen) {
     do {
 
@@ -760,17 +905,17 @@ int pr_netio_write_async(pr_netio_stream_t *nstrm, char *buf, size_t buflen) {
       switch (nstrm->strm_type) {
         case PR_NETIO_STRM_CTRL:
           bwritten = ctrl_netio ? (ctrl_netio->write)(nstrm, buf, buflen) :
-            (core_ctrl_netio->write)(nstrm, buf, buflen);
+            (default_ctrl_netio->write)(nstrm, buf, buflen);
           break;
 
         case PR_NETIO_STRM_DATA:
           bwritten = data_netio ? (data_netio->write)(nstrm, buf, buflen) :
-            (core_data_netio->write)(nstrm, buf, buflen);
+            (default_data_netio->write)(nstrm, buf, buflen);
           break;
 
         case PR_NETIO_STRM_OTHR:
           bwritten = othr_netio ? (othr_netio->write)(nstrm, buf, buflen) :
-            (core_othr_netio->write)(nstrm, buf, buflen);
+            (default_othr_netio->write)(nstrm, buf, buflen);
           break;
       }
 
@@ -836,7 +981,7 @@ int pr_netio_read(pr_netio_stream_t *nstrm, char *buf, size_t buflen,
           switch (nstrm->strm_type) {
             case PR_NETIO_STRM_CTRL:
               bread = ctrl_netio ? (ctrl_netio->read)(nstrm, buf, buflen) :
-                (core_ctrl_netio->read)(nstrm, buf, buflen);
+                (default_ctrl_netio->read)(nstrm, buf, buflen);
                 break;
 
             case PR_NETIO_STRM_DATA:
@@ -844,18 +989,34 @@ int pr_netio_read(pr_netio_stream_t *nstrm, char *buf, size_t buflen,
                 break;
 
               bread = data_netio ? (data_netio->read)(nstrm, buf, buflen) :
-                (core_data_netio->read)(nstrm, buf, buflen);
+                (default_data_netio->read)(nstrm, buf, buflen);
               break;
 
             case PR_NETIO_STRM_OTHR:
               bread = othr_netio ? (othr_netio->read)(nstrm, buf, buflen) :
-                (core_othr_netio->read)(nstrm, buf, buflen);
+                (default_othr_netio->read)(nstrm, buf, buflen);
               break;
           }
 
 #ifdef EAGAIN
-	  if (bread == -1 && errno == EAGAIN)
+	  if (bread == -1 &&
+              errno == EAGAIN) {
+            int xerrno = EAGAIN;
+
+            /* Treat this as an interrupted call, call pr_signals_handle()
+             * (which will delay for a few msecs because of EINTR), and try
+             * again.
+             *
+             * This should avoid a tightly spinning loop if read(2) returns
+             * EAGAIN, as on a data transfer (Bug#3639).
+             */
+
+            errno = EINTR;
+            pr_signals_handle();
+
+            errno = xerrno;
             goto polling;
+          }
 #endif
 
         } while (bread == -1 && errno == EINTR);
@@ -885,6 +1046,7 @@ int pr_netio_read(pr_netio_stream_t *nstrm, char *buf, size_t buflen,
     buflen -= bread;
   }
 
+  session.total_raw_in += total;
   return total;
 }
 
@@ -896,22 +1058,21 @@ int pr_netio_shutdown(pr_netio_stream_t *nstrm, int how) {
     return -1;
   }
 
-  if (nstrm->strm_type == PR_NETIO_STRM_CTRL) {
-    res = ctrl_netio ? (ctrl_netio->shutdown)(nstrm, how) :
-      (core_ctrl_netio->shutdown)(nstrm, how);
-    return res;
-  }
+  switch (nstrm->strm_type) {
+    case PR_NETIO_STRM_CTRL:
+      res = ctrl_netio ? (ctrl_netio->shutdown)(nstrm, how) :
+        (default_ctrl_netio->shutdown)(nstrm, how);
+      return res;
 
-  if (nstrm->strm_type == PR_NETIO_STRM_DATA) {
-    res = data_netio ? (data_netio->shutdown)(nstrm, how) :
-      (core_data_netio->shutdown)(nstrm, how);
-    return res;
-  }
+    case PR_NETIO_STRM_DATA:
+      res = data_netio ? (data_netio->shutdown)(nstrm, how) :
+        (default_data_netio->shutdown)(nstrm, how);
+      return res;
 
-  if (nstrm->strm_type == PR_NETIO_STRM_OTHR) {
-    res = othr_netio ? (othr_netio->shutdown)(nstrm, how) :
-      (core_othr_netio->shutdown)(nstrm, how);
-    return res;
+    case PR_NETIO_STRM_OTHR:
+      res = othr_netio ? (othr_netio->shutdown)(nstrm, how) :
+        (default_othr_netio->shutdown)(nstrm, how);
+      return res;
   }
 
   errno = EPERM;
@@ -930,10 +1091,12 @@ char *pr_netio_gets(char *buf, size_t buflen, pr_netio_stream_t *nstrm) {
 
   buflen--;
 
-  if (nstrm->strm_buf)
+  if (nstrm->strm_buf) {
     pbuf = nstrm->strm_buf;
-  else
-    pbuf = netio_buffer_alloc(nstrm);
+
+  } else {
+    pbuf = pr_netio_buffer_alloc(nstrm);
+  }
 
   while (buflen) {
 
@@ -956,8 +1119,17 @@ char *pr_netio_gets(char *buf, size_t buflen, pr_netio_stream_t *nstrm) {
       pbuf->remaining = pbuf->buflen - toread;
       pbuf->current = pbuf->buf;
 
-    } else
-      toread = pbuf->buflen - pbuf->remaining;
+      pbuf->remaining = pbuf->buflen - toread;
+      pbuf->current = pbuf->buf;
+
+      /* Before we begin iterating through the data read in from the
+       * network, generate an event for any listeners which may want to
+       * examine this data as well.
+       */
+      pr_event_generate("core.othr-read", pbuf);
+    }
+
+    toread = pbuf->buflen - pbuf->remaining;
 
     while (buflen && *pbuf->current != '\n' && toread--) {
       if (*pbuf->current & 0x80)
@@ -995,7 +1167,9 @@ char *pr_netio_telnet_gets(char *buf, size_t buflen,
   int toread, handle_iac = TRUE, saw_newline = FALSE;
   pr_buffer_t *pbuf = NULL;
 
-  if (buflen == 0) {
+  if (buflen == 0 ||
+      in_nstrm == NULL ||
+      out_nstrm == NULL) {
     errno = EINVAL;
     return NULL;
   }
@@ -1006,37 +1180,50 @@ char *pr_netio_telnet_gets(char *buf, size_t buflen,
 
   buflen--;
 
-  if (in_nstrm->strm_buf)
+  if (in_nstrm->strm_buf) {
     pbuf = in_nstrm->strm_buf;
-  else
-    pbuf = netio_buffer_alloc(in_nstrm);
 
-  while (buflen) {
+  } else {
+    pbuf = pr_netio_buffer_alloc(in_nstrm);
+  }
+
+  while (buflen > 0) {
+    pr_signals_handle();
 
     /* Is the buffer empty? */
-    if (!pbuf->current ||
+    if (pbuf->current == NULL ||
         pbuf->remaining == pbuf->buflen) {
 
       toread = pr_netio_read(in_nstrm, pbuf->buf,
-        (buflen < pbuf->buflen ?  buflen : pbuf->buflen), 1);
+        (buflen < pbuf->buflen ? buflen : pbuf->buflen), 1);
 
       if (toread <= 0) {
         if (bp != buf) {
           *bp = '\0';
           return buf;
-
-        } else {
-          return NULL;
         }
+
+        return NULL;
       }
 
       pbuf->remaining = pbuf->buflen - toread;
       pbuf->current = pbuf->buf;
 
-    } else
-      toread = pbuf->buflen - pbuf->remaining;
+      /* Before we begin iterating through the data read in from the
+       * network, handing any Telnet characters and such, generate an event
+       * for any listeners which may want to examine this data as well.
+       */
+      pr_event_generate("core.ctrl-read", pbuf);
+    }
+
+    toread = pbuf->buflen - pbuf->remaining;
+
+    while (buflen > 0 &&
+           toread > 0 &&
+           *pbuf->current != '\n' &&
+           toread--) {
+      pr_signals_handle();
 
-    while (buflen && toread > 0 && *pbuf->current != '\n' && toread--) {
       cp = *pbuf->current++;
       pbuf->remaining++;
 
@@ -1050,7 +1237,6 @@ char *pr_netio_telnet_gets(char *buf, size_t buflen,
               case TELNET_DONT:
               case TELNET_IP:
               case TELNET_DM:
-
                 /* Why do we do this crazy thing where we set the "telnet mode"
                  * to be the action, and let the while loop, on the next pass,
                  * handle that action?  It's because we don't know, right now,
@@ -1062,6 +1248,16 @@ char *pr_netio_telnet_gets(char *buf, size_t buflen,
                 telnet_mode = cp;
                 continue;
 
+              case TELNET_IAC:
+                /* In this case, we know that the previous byte was TELNET_IAC,
+                 * and that the current byte is another TELNET_IAC.  The
+                 * first TELNET_IAC thus "escapes" the second, telling us
+                 * that the current byte (TELNET_IAC) should be written out
+                 * as is (Bug#3697).
+                 */
+                telnet_mode = 0;
+                break;
+
               default:
                 /* In this case, we know that the previous byte was TELNET_IAC,
                  * but the current byte is not a value we care about.  So
@@ -1114,7 +1310,9 @@ char *pr_netio_telnet_gets(char *buf, size_t buflen,
       buflen--;
     }
 
-    if (buflen && toread && *pbuf->current == '\n') {
+    if (buflen > 0 &&
+        toread > 0 &&
+        *pbuf->current == '\n') {
       buflen--;
       toread--;
       *bp++ = *pbuf->current++;
@@ -1124,8 +1322,12 @@ char *pr_netio_telnet_gets(char *buf, size_t buflen,
       break;
     }
 
-    if (!toread)
+    if (toread == 0) {
+      /* No more input?  Set pbuf->current to null, so that at the top of
+       * the loop, we read more.
+       */
       pbuf->current = NULL;
+    }
   }
 
   if (!saw_newline) {
@@ -1153,21 +1355,26 @@ char *pr_netio_telnet_gets(char *buf, size_t buflen,
 
 int pr_register_netio(pr_netio_t *netio, int strm_types) {
 
-  if (!netio) {
-    pr_netio_t *core_netio = NULL;
+  if (netio == NULL) {
+    pr_netio_t *default_netio = NULL;
 
-    /* Only instantiate the core NetIO objects once, reusing the same pointer.
+    /* Only instantiate the default NetIO objects once, reusing the same
+     * pointer.
      */
-    if (!core_ctrl_netio)
-      core_netio = core_ctrl_netio = pr_alloc_netio(permanent_pool);
+    if (default_ctrl_netio == NULL) {
+      default_ctrl_netio = default_netio = pr_alloc_netio2(permanent_pool,
+        NULL);
+    }
 
-    if (!core_data_netio)
-      core_data_netio = core_netio ? core_netio :
-        (core_netio = pr_alloc_netio(permanent_pool));
+    if (default_data_netio == NULL) {
+      default_data_netio = default_netio ? default_netio :
+        (default_netio = pr_alloc_netio2(permanent_pool, NULL));
+    }
 
-    if (!core_othr_netio)
-      core_othr_netio = core_netio ? core_netio :
-        (core_netio = pr_alloc_netio(permanent_pool));
+    if (default_othr_netio == NULL) {
+      default_othr_netio = default_netio ? default_netio :
+        (default_netio = pr_alloc_netio2(permanent_pool, NULL));
+    }
 
     return 0;
   }
@@ -1179,20 +1386,22 @@ int pr_register_netio(pr_netio_t *netio, int strm_types) {
     return -1;
   }
 
-  if (strm_types & PR_NETIO_STRM_CTRL)
+  if (strm_types & PR_NETIO_STRM_CTRL) {
     ctrl_netio = netio;
+  }
 
-  if (strm_types & PR_NETIO_STRM_DATA)
+  if (strm_types & PR_NETIO_STRM_DATA) {
     data_netio = netio;
+  }
 
-  if (strm_types & PR_NETIO_STRM_OTHR)
+  if (strm_types & PR_NETIO_STRM_OTHR) {
     othr_netio = netio;
+  }
 
   return 0;
 }
 
 int pr_unregister_netio(int strm_types) {
-
   if (!strm_types) {
     errno = EINVAL;
     return -1;
@@ -1200,25 +1409,56 @@ int pr_unregister_netio(int strm_types) {
 
   /* NOTE: consider using cleanups here in the future? */
 
-  if (strm_types & PR_NETIO_STRM_CTRL)
+  if (strm_types & PR_NETIO_STRM_CTRL) {
     ctrl_netio = NULL;
+  }
 
-  if (strm_types & PR_NETIO_STRM_DATA)
+  if (strm_types & PR_NETIO_STRM_DATA) {
     data_netio = NULL;
+  }
 
-  if (strm_types & PR_NETIO_STRM_OTHR)
+  if (strm_types & PR_NETIO_STRM_OTHR) {
     othr_netio = NULL;
+  }
 
   return 0;
 }
 
+pr_netio_t *pr_get_netio(int strm_type) {
+  pr_netio_t *netio = NULL;
+
+  if (strm_type == 0) {
+    errno = EINVAL;
+    return NULL;
+  }
+
+  switch (strm_type) {
+    case PR_NETIO_STRM_CTRL:
+      netio = ctrl_netio;
+      break;
+
+    case PR_NETIO_STRM_DATA:
+      netio = data_netio;
+      break;
+
+    case PR_NETIO_STRM_OTHR:
+      netio = othr_netio;
+      break;
+
+    default:
+      errno = ENOENT;
+  }
+
+  return netio;
+}
+
 extern pid_t mpid;
 
-pr_netio_t *pr_alloc_netio(pool *parent_pool) {
+pr_netio_t *pr_alloc_netio2(pool *parent_pool, module *owner) {
   pr_netio_t *netio = NULL;
   pool *netio_pool = NULL;
 
-  if (!parent_pool) {
+  if (parent_pool == NULL) {
     errno = EINVAL;
     return NULL;
   }
@@ -1242,6 +1482,11 @@ pr_netio_t *pr_alloc_netio(pool *parent_pool) {
 
   netio = pcalloc(netio_pool, sizeof(pr_netio_t));
   netio->pool = netio_pool;
+  netio->owner = owner;
+
+  if (owner != NULL) {
+    netio->owner_name = pstrdup(netio_pool, owner->name);
+  }
 
   /* Set the default NetIO handlers to the core handlers. */
   netio->abort = core_netio_abort_cb;
@@ -1257,6 +1502,10 @@ pr_netio_t *pr_alloc_netio(pool *parent_pool) {
   return netio;
 }
 
+pr_netio_t *pr_alloc_netio(pool *parent_pool) {
+  return pr_alloc_netio2(parent_pool, NULL);
+}
+
 void init_netio(void) {
   signal(SIGPIPE, SIG_IGN);
   signal(SIGURG, SIG_IGN);
